BeginPackage["HEPMath`Lorentz`", {"HEPMath`Tensors`"}]


(* index types *)

Lorentz::usage = "Lorentz is the index type for Lorentz indices"

Dirac::usage = "Dirac is the index type for Dirac indices"

FermionHelicity::usage = "FermionHelicity is the index type for fermion helicity
  indices."

VectorPolarization::usage = "VectorPolarization is the index type for massless
vector boson polarizations."

MassiveVectorPolarization::usage = "MassiveVectorPolarization is the index type
for massive vector boson polarizations."


(* declarations *)

DeclareLorentzVectors::usage = "DeclareLorentzVectors[p, q, ...] is equivalent
to DeclareRealHEPTensor[{p, q, ...}, {Lorentz}]."

DeclareComplexLorentzVectors::usage = "DeclareComplexLorentzVectors[p, q, ...]
is equivalent to DeclareHEPTensor[{p, q, ...}, {Lorentz}]."

LorentzVectorQ::usage = "LorentzVectorQ[expr] is equivalent to
HEPTensorSignature[expr] === {Lorentz}."

DeclareDiracSpinors::usage = "DeclareDiracSpinors[u, v, ...] is equivalent
to DeclareHEPTensor[{p, q, ...}, {Dirac}]."

DiracSpinorQ::usage = "DiracSpinorQ[expr] is equivalent to
HEPTensorSignature[expr] === {Dirac}."

DeclareDiracMatrices::usage = "DeclareDiracMatrices[M1, M2, ...] is equivalent
to DeclareHEPTensor[{M1, M2, ...}, {Dirac, Dirac}]."

DiracMatrixQ::usage = "DiracMatrixQ[expr] is equivalent to
HEPTensorSignature[expr] === {Dirac, Dirac}."


(* metric and Levi-Civita tensors *)

Eta::usage = "Eta represents the metic tensor. Its tensor signature is
 {Lorentz, Lorentz}."

Eps::usage = "Eps represents the four-dimensional Levi-Civita tensor. Its
tensor signature is {Lorentz, Lorentz, Lorentz, Lorentz}."

EpsDot::usage = "EpsDot[p, q, ...] denotes the contraction of Lorentz vectors
p, q, ... with the Levi-Civita tensor. Depending on the number of arguments,
the tensor signature of EpsDot[p, ...] can consist of zero to three Lorentz
indices. The Lorentz vectors are contracted with the first indices of Eps.
For example, EpsDot[p, q][mu, nu] is equivalent to Eps[ro, sg, mu, nu] *
p[ro] q[sg]."

EpsDot::invarg = "Invalid argument to EpsDot in `1`"

EpsDot::toomany = "Too many arguments to EpsDot: `1`"

MProd::usage = "MProd[p, q] denotes the Minkowski product of p and q. The
arguments p and q can be expressions with tensor signature {Lorentz} or
lists with four elements. In the latter case, MProd[p, q] evaluates to
p[[1]] q[[1]] - p[[2]] q[[2]] - p[[3]] q[[3]] - p[[4]] q[[4]]."


(* Dirac matrices *)

Ga::usage = "Ga is the tensor representing Dirac gamma matrices. Its signature
is {Lorentz, Dirac, Dirac}"

Gs::usage = "Gs[p] is the tensor representing the contraction of the Lorentz
vector p with the Dirac gamma matrices. Its signature is {Dirac, Dirac}. The
signature of p must be {Lorentz}."

Gs::invarg = "Invalid argument to Gs: `1`"

GI::usage = "GI is the tensor representing the unit Dirac matrix. Its signature
is {Dirac, Dirac}."

G5::usage = "G5 is the tensor representing the Dirac gamma_5 matrix. Its
signature is {Dirac, Dirac}."

G0::usage = "G5 is the tensor representing the Dirac gamma_0 matrix. Its
signature is {Dirac, Dirac}."

PL::usage = "PL is the tensor representing the left-handed Dirac projection
matrix PL=(G1-G5)/2. Its signature is {Dirac, Dirac}."

PR::usage = "PR is the tensor representing the right-handed Dirac projection
matrix PR=(G1+G5)/2. Its signature is {Dirac, Dirac}."

GSig::usage = "GSig is the tensor representing the generators of
Lorentz transformations on the space of Dirac spinors:
GSig[mu,nu,a,b] = I/2 (Ga[mu].Ga[nu] - Ga[nu].Ga[mu]). Its signature
is {Lorentz, Lorentz, Dirac, Dirac}."


(* Dirac spinors *)

USp::usage = "USp[p] is the tensor representing the Dirac u spinor for the
four-momentum p. Its tensor signature is {FermionHelicity, Dirac}. The
signature of p must be {Lorentz}."

USp::invarg = "Invalid argument to USp: `1`"

VSp::usage = "VSp[p] is the tensor representing the Dirac v spinor for the
four-momentum p. Its tensor signature is {FermionHelicity, Dirac}. The
signature of p must be {Lorentz}."

VSp::invarg = "Invalid argument to VSp: `1`"

Bar::usage = "Bar[expr] denotes the bar conjugate of expression expr."


(* Polarization vectors *)

Pol::usage = "Pol[p] represents the polarization vector of a massless vector
boson with four momentum p. Its tensor signature is {VectorPolarization, 
Lorentz}. The signature of p must be {Lorentz}."

MPol::usage = "MPol[p] represents the polarization vector of a massive vector
boson with four momentum p. Its tensor signature is {MassiveVectorPolarization, 
Lorentz}. The signature of p must be {Lorentz}."


(* contractions *)

HDot::usage = "HDot[p,q] represents the product of the (D-4)-dimensional
parts of the Lorentz vectors p and q."

DiracTr::usage = "DiracTr[Gamma] represents the trace of the Dirac matrix Gamma."


(* manipulations *)

DiracSubstitute::usage = "DiracSubstitute[expr] replaces occurrences of GSig,
PL and PR in expr by their definitions."

DiracExpand::usage = "DiracExpand[expr] expands expr with respect to Dirac
spinors and matrices. It is equivalent to DiracContract[HEPExpand[
  DiracSubstitute[expr], Gs|Ga|GI|G5]]"

DiracContract::usage = "DiracContract[expr] contracts all Dirac indices in
expr, using index suppressed notation (Dot) and DiracTr wherever possible and
removing spurious occurrences of GI."

ExplicitSlash::usage = "ExplicitSlash[expr] replaces subexpressions in expr
of the form Gs[p] by explicit index notation Ga[mu] p[mu]. Unique index
symbols are generated automatically. ExplicitSlash[expr, patt] only applies
to Lorentz vectors p matching pattern patt."

SortEps::usage = "SortEps[exrp] uses anti-symmetry to bring Eps and EpsDot
expressions in expr in canonical order."

LorentzContract::usage = "LorentzContract[expr] contracts all Lorentz indices
in expr, eliminating Eta, EtaHat and Eps tensors where possible."

PullVectors::usage = "PullVectors[expr, patt] pulls Lorentz vectors matching
patt out of contractions such as Gs, DiracTr or dotted chains of Dirac
matrices, introducing dummy indices where necessary."

ToComponents::usage = "ToComponents[expr, {p -> {p0, p1, p2, p3},
k -> {k0, k1, k2, k3}, ...}] expresses scalar products of the Lorentz vectors
p, k, etc in expr in terms of their components, p0, p1, ..., k0, k1, etc."

Options[ToComponents] = {PostProcess -> Identity}


(* traces *)

Dim::usage = "Dim is the default dimension of Lorentz vectors"

FourDimensionalQ::usage = "FourDimensionalQ[p] returns True if the Lorentz
vector p has been declared four dimensional."

DeclareFourDimensional::usage = "DeclareFourDimensional[p, q, ...] declares
the Lorentz vectors p, q, ... as four dimensional."

Declare4DLorentzVectors::usage = "Declare4DLorentzVectors[p, q, ...] is
equivalent to DeclareRealHEPTensor[{p, q, ...}, {Lorentz}];
DeclareFourDimensional[p, q, ...]."

DeclareComplex4DLorentzVectors::usage =
"DeclareComplex4DLorentzVectors[p, q, ...] is equivalent to
DeclareHEPTensor[{p, q, ...}, {Lorentz}]; DeclareFourDimensional[p, q, ...]."

EtaHat::usage = "EtaHat represents the (D-4)-dimensional part of the metric
tensor. Its tensor signature is {Lorentz, Lorentz}."

CalcDiracTraces::usage = "CalcDiracTraces[expr] calculates all Dirac traces
in expr"

Options[CalcDiracTraces] = {AntiCommutingG5 -> False}

AntiCommutingG5::usage = "AntiCommutingG5 is an option for CalcDiracTraces. If
set to True, traces are computed with an anti-commuting gamma_5. Defaults to
False."


(* Polarization sums *)

FermionPolarizationSums::usage =
"FermionPolarizationSums[expr, {p1->m1, p2->m2, ...}] replaces sums over fermion
polarizations in expr by the corresponding Dirac matrices, using mass m1 for the
fermion with momentum p1, mass m2 for the fermion with momentum p2 etc."

FermionPolarizationSums::invm = "Invalid mass for external momentum `1`."

VectorPolarizationSums::usage =
"VectorPolarizationSums[expr, {p->m, ..., k->n, ...}] replaces sums over vector
boson polarizations in expr by the corresponding projectors, using mass m for
the massive vector boson with momentum p and gauge vector n for the massless
vector boson with momentum k."

VectorPolarizationSums::invm = "Invalid mass for external momentum `1`."

VectorPolarizationSums::invn = "Invalid gauge vector for external momentum `1`."

PostProcess::usage = "PostProcess is an option of VectorPolarizationSums
and ToComponents. In the case of VectorPolarisationSums the value of
PostProcess is applied to denominators in the polarization sum of massless vector
bosons. In the case of ToComponents it is applied to the scalar products after
substituting the vector components. It defaults ot Identity."

Options[VectorPolarizationSums] = {PostProcess -> Identity}


Begin["`Private`"]

tilde::usage = "tilde is a private symbol which indicates a 4D index"

hat::usage = "tilde is a private symbol which indicates a D-4 index"

g5::usage = "g5 is a private symbol denoting gamma_5"

tr::usage = "tr is the internal function which computes the trace."

contractme::usage = "contractme marks an expression to be LorentzContract-ed."

antiCommutingG5::usage = "internal symbol idicating anti-commuting gamma_5."

tensorQ::usage = "test if an expression is a tensor."

tensorQ[ex_] := HEPTensorSignature[ex] =!= {}

(* index types *)

DeclareIndexType[Lorentz, Dim, Eta]

DeclareIndexType[Dirac, 4, GI]

DeclareIndexType[FermionHelicity, 2]

DeclareIndexType[VectorPolarization, 2]

DeclareIndexType[MassiveVectorPolarization, 3]


(* declarations *)

DeclareLorentzVectors[aa__] := DeclareRealHEPTensor[{aa}, {Lorentz}]

DeclareComplexLorentzVectors[aa__] := DeclareHEPTensor[{aa}, {Lorentz}]

LorentzVectorQ[x_] := HEPTensorSignature[x] === {Lorentz}

DeclareDiracSpinors[aa__] := DeclareHEPTensor[{aa}, {Dirac}]

DiracSpinorQ[x_] := HEPTensorSignature[x] === {Dirac}

DeclareDiracMatrices[aa__] := DeclareHEPTensor[{aa}, {Dirac, Dirac}]

DiracMatrixQ[x_] := HEPTensorSignature[x] === {Dirac, Dirac}

DeclareFourDimensional[s__] := (FourDimensionalQ[#] = True)& /@ {s}

Declare4DLorentzVectors[aa__] := (
  DeclareLorentzVectors[aa]; DeclareFourDimensional[aa])

DeclareComplex4DLorentzVectors[aa__] := (
  DeclareComplexLorentzVectors[aa]; DeclareFourDimensional[aa])


ScopeSymbols[FourDimensionalQ]

FourDimensionalQ[s_Plus] := And @@ (FourDimensionalQ /@ (List @@ s))

FourDimensionalQ[a_ p_?LorentzVectorQ] := FourDimensionalQ[p]


(* Levi-Civita tensor *)

DeclareRealHEPTensor[Eps, {Lorentz, Lorentz, Lorentz, Lorentz}]

DeclareLinear[EpsDot]

DeclareRealHolomorphic[EpsDot]

DeclareHEPTensor[EpsDot[_], {Lorentz, Lorentz, Lorentz},
  If[(HEPTensorSignature /@ (List @@ #)) =!= {{Lorentz}},
     Message[EpsDot::invarg, #]; Abort[]]&
]

DeclareHEPTensor[EpsDot[_, _], {Lorentz, Lorentz},
  If[(HEPTensorSignature /@ (List @@ #)) =!= {{Lorentz}, {Lorentz}},
     Message[EpsDot::invarg, #]; Abort[]]&
]

DeclareHEPTensor[EpsDot[_, _, _], {Lorentz},
  If[(HEPTensorSignature /@ (List @@ #)) =!= {{Lorentz}, {Lorentz}, {Lorentz}},
     Message[EpsDot::invarg, #]; Abort[]]&
]

DeclareHEPTensor[EpsDot[_, _, _, _], {},
  If[(HEPTensorSignature /@ (List @@ #)) =!=
      {{Lorentz}, {Lorentz}, {Lorentz}, {Lorentz}},
     Message[EpsDot::invarg, #]; Abort[]]&
]

EpsDot[a_, b_, c_, d_, ee__] :=
  (Message[EpsDot::toomany, {a, b, c, d, ee}]; Abort[])

EpsDot[] = Eps

UserExplicitIndices[e:EpsDot[pp__][ii__], patt_] := Block[{jj},
  jj = Table[Unique[], {Length[{pp}]}];
  (Eps @@ Join[jj, {ii}]) (Times @@ MapThread[#1[#2]&, {{pp}, jj}])
] /; !FreeQ[e, patt]



(* SortEps *)

SortEps[ex_] := Block[{res=ex},
  res = res /. (Eps|(_EpsDot))[___, a_, ___, a_, ___] -> 0;
  res = res /. HoldPattern[EpsDot[___, p_, ___, p_, ___][__]] -> 0;
  res = res /. HoldPattern[EpsDot[___, p_, ___, p_, ___]] -> 0;

  res = res /. HoldPattern[Eps[mm__]] :> Signature[{mm}] (Eps @@ Sort[{mm}]);
  res = res /. HoldPattern[EpsDot[pp__][mm__]] :>
    Signature[{pp}] Signature[{mm}] ((EpsDot @@ Sort[{pp}]) @@ Sort[{mm}]);
  res = res /. HoldPattern[EpsDot[pp__]] :>
    Signature[{pp}] (EpsDot @@ Sort[{pp}]);
  
  res
]


(* metric tensor *)

Attributes[EtaHat] = {Orderless}

DeclareRealHEPTensor[EtaHat, {Lorentz, Lorentz}]


(* MProd *)

SetAttributes[MProd, Orderless]

DeclareLinear[MProd]

DeclareRealHolomorphic[MProd]

MProd[{p0_, px_, py_, pz_}, {q0_, qx_, qy_, qz_}] =
  p0 q0 - px qx - py qy - pz qz


(* Dirac Matrices *)

DeclareHEPTensor[Ga, {Lorentz, Dirac, Dirac}]

DeclareHEPTensor[Gs[_], {Dirac, Dirac},
  If[HEPTensorSignature[First[#]] =!= {Lorentz},
     Message[Gs::invarg, First[#]]; Abort[]]&]

DeclareLinear[Gs]

DeclareHEPTensor[{G5, G0, PL, PR}, {Dirac, Dirac}]

DeclareHEPTensor[GSig, {Lorentz, Lorentz, Dirac, Dirac}]


(* Dirac spinors and bar conjugation *)

DeclareHEPTensor[USp[_], {FermionHelicity, Dirac},
  If[HEPTensorSignature[First[#]] =!= {Lorentz},
     Message[USp::invarg, First[#]]; Abort[]]&]

DeclareHEPTensor[VSp[_], {FermionHelicity, Dirac},
  If[HEPTensorSignature[First[#]] =!= {Lorentz},
     Message[VSp::invarg, First[#]]; Abort[]]&]

noDiracQ[x_] := FreeQ[HEPTensorSignature[x], Dirac]

DeclareLinear[Bar]

HEPTensorSignature[Bar[x_]] := HEPTensorSignature[x]

Bar[s_Plus] := Bar /@ s

Bar[p_Times] := Bar /@ p

Bar[Bar[x_]] = x

Bar[(u_?DiracSpinorQ).(v_?DiracSpinorQ)] := Bar[v].Bar[u]

Bar[(u_?DiracSpinorQ).(g_?DiracMatrixQ)] := Bar[g].Bar[u]

Bar[(g_?DiracMatrixQ).(u_?DiracSpinorQ)] := Bar[u].Bar[g]

Bar[(g1_?DiracMatrixQ).(g2_?DiracMatrixQ)] := Bar[g2].Bar[g1]

Bar[a_?noDiracQ] := CC[a]

Bar[Ga[mu_]] = Ga[mu]

Bar[Gs[p_]] = Gs[CC[p]]

Bar[GI] = GI

Bar[G5] = -G5

Bar[G0] = G0

Bar[PL] = PR

Bar[PR] = PL

Bar[GSig[mu_,nu_]] = GSig[mu,nu]

Bar[USp[p_][ii__]] := Bar[USp[p]][ii]

Bar[VSp[p_][ii__]] := Bar[VSp[p]][ii]

Bar[HoldPattern[Bar[USp[p_]][ii__]]] := USp[p][ii]

Bar[HoldPattern[Bar[VSp[p_]][ii__]]] := VSp[p][ii]

DownValues[CC] = Join[{HoldPattern[CC[(u_?DiracSpinorQ).(v_?DiracSpinorQ)]] :>
  Bar[u.v]}, DownValues[CC]]


(* Polarization vectors and complex conjugation *)

DeclareHEPTensor[Pol[_], {VectorPolarization, Lorentz},
  If[HEPTensorSignature[First[#]] =!= {Lorentz},
     Message[USp::invarg, First[#]]; Abort[]]&]

DeclareHEPTensor[MPol[_], {MassiveVectorPolarization, Lorentz},
  If[HEPTensorSignature[First[#]] =!= {Lorentz},
     Message[VSp::invarg, First[#]]; Abort[]]&]

CC[Pol[p_][ii__]] := CC[Pol[p]][ii]

CC[MPol[p_][ii__]] := CC[MPol[p]][ii]


(* contractions *)

Attributes[HDot] = {Orderless}

DeclareLinear[HDot]

DeclareRealHolomorphic[HDot]

DeclareLinear[DiracTr]

CC[DiracTr[HoldPattern[Dot[aa__]]]] := DiracTr[Dot@@(Bar /@ Reverse[{aa}])]

CC[DiracTr[a_]] := DiracTr[Bar[a]]

UserExplicitIndices[e:DiracTr[g_], patt_] := Block[{al=Unique[]},
  ExplicitIndices[g[al, al], patt]] /; !FreeQ[e, patt]


(* DiracSubstitute *)

DiracSubstitute[ex_] := ex /. {
  GSig[mu_, nu_] -> (I/2) (Ga[mu].Ga[nu] - Ga[nu].Ga[mu]),
  PL -> (GI - G5)/2,
  PR -> (GI + G5)/2
}


(* DiracContract *)

isDiracMatrix[t_, l_] := Block[{sig=HEPTensorSignature[t]},
  Length[sig] === Length[l]+2 && Take[sig, -2] === {Dirac, Dirac}]

isDiracSpinor[u_, l_] := Block[{sig=HEPTensorSignature[u]},
  Length[sig] === Length[l]+1 && Last[sig] === Dirac]

partSub[t_, {}] := t

partSub[t_, {ii__}] := t[ii]

DiracContract[ex_] := HEPContract[ex, GI] //. {
  (t1_[ii___, al_, bt_] /; isDiracMatrix[t1, {ii}]) *
  (t2_[jj___, bt_, gm_] /; isDiracMatrix[t2, {jj}]) :>
    (partSub[t1, {ii}].partSub[t2, {jj}])[al,gm],

  (t_[ii___, al_, bt_] /; isDiracMatrix[t, {ii}]) *
  (u_[jj___, bt_] /; isDiracSpinor[u, {jj}]) :>
    (partSub[t, {ii}].partSub[u, {jj}])[al],

  (u_[ii___, al_] /; isDiracSpinor[u,{ii}]) *
  (t_[jj___, al_, bt_] /; isDiracMatrix[t, {jj}]) :>
    (partSub[u, {ii}].partSub[t, {jj}])[bt],

  (u_[ii___, al_] /; isDiracSpinor[u,{ii}]) *
  (v_[jj___, al_] /; isDiracSpinor[v,{jj}]) :>
    partSub[u, {ii}].partSub[v, {jj}],

  (t_[ii___, al_, al_] /; isDiracMatrix[t,{ii}]) :>
    DiracTr[partSub[t, {ii}]]
} /. Ga[mu_][al_,bt_] -> Ga[mu,al,bt]


(* DiracExpand *)

DiracExpand[ex_] := DiracContract[HEPExpand[DiracSubstitute[ex], Gs|Ga|GI|G5]]


(* ExplicitSlash *)

ExplicitSlash[ex_, patt_] :=
  JoinIndexLists[ex /. Gs[p:patt] :> Block[{mu=Unique[]}, p[mu] Ga[mu]]]
             
ExplicitSlash[ex_] := ExplicitSlash[ex, _]


(* LorentzContract *)

eps::usage = "eps is an internal symbol representing the Levi-Civita tensor."

ind::usage = "ind flags an index in an argument list"

epsOrder::usage = "epsOrder is the order function for eps argument lists."

epsOrder[ind[a_], ind[b_]] := OrderedQ[{a, b}]

epsOrder[ind[_], _] = False

epsOrder[_, ind[_]] = True

epsOrder[a_, b_] := OrderedQ[{a, b}]


LorentzContract[ex_] := Block[{res=ex},
  res = HEPContract[res, Eta];
  If[!FreeQ[res, Eps|EpsDot],
    res = res /. HoldPattern[Eps[___, a_, ___, a_, ___]] -> 0;
    res = res /. HoldPattern[EpsDot[___][___, a_, ___, a_, ___]] -> 0;
    res = res /.
      HoldPattern[EtaHat[___,a_,___]] HoldPattern[Eps[___, a_, ___]] -> 0;
    res = res /.
      HoldPattern[EtaHat[___,a_,___]] HoldPattern[EpsDot[___][___, a_, ___]]->0;
    res = res /.
      HoldPattern[Eps[___, a_, ___, b_, ___]] *
        (t_?tensorQ)[x___,a_,y___] t_[x___,b_,y___] -> 0;
    res = res /.
      HoldPattern[EpsDot[pp___][___, a_, ___, b_, ___]] *
        (t_?tensorQ)[x___,a_,y___] t_[x___,b_,y___] -> 0;

    res = res /. Eps[ii___] :> eps @@ (ind /@ {ii});
    res = res /. EpsDot[pp___][ii___] :> eps @@ Join[{pp}, ind /@ {ii}];
    res = res //. HoldPattern[eps[a___, ind[i_], b___]] p_[i_] :> eps[a, p, b];
    res = res /. eps[aa___] :> Block[{ord},
      ord = Sort[{aa}, epsOrder];
      Signature[ord] Signature[{aa}] (eps @@ ord)];
    res = res /. eps[pp___, ii:Longest[___ind]] :>
      (eps[pp][ii] /. ind -> Identity);
    res = res /. eps[][ii__] :> Eps[ii];
    res = res /. eps[pp__][] :> EpsDot[pp];
    res = res /. eps[pp__][ii__] :> EpsDot[pp][ii];
  ];

  res = res //. HoldPattern[(h:(Eta|EtaHat))[p1___,a_,p2___]] EtaHat[b_,a_] ->
    EtaHat[p1,b,p2];
  res = res /. EtaHat[mu_, nu_] (p_?FourDimensionalQ)[nu_] -> 0;
  res = res //. Eta[mu_, nu_] p_[mu_] q_[nu_] -> SDot[p, q];
  res = res //. EtaHat[mu_, nu_] p_[mu_] q_[nu_] -> HDot[p, q];
  res = res /. {
    EtaHat[mu_, nu_]^2 -> (Dim-4),
    EtaHat[mu_, mu_] -> (Dim-4)
  };
  res = res //. p_[mu_] q_[mu_] :> SDot[p,q] /;
    HEPTensorSignature[p] === HEPTensorSignature[q] === {Lorentz};
  res = res /. p_[mu_]^2 :> Sqr[p] /; HEPTensorSignature[p] === {Lorentz};

  res
]


(* PullVectors *)

pullVectors[x_, patt_] := x /; FreeQ[x, patt]

pullVectors[s_Plus, patt_] := pullVectors[#, patt]& /@ s

pullVectors[p_Times, patt_] := pullVectors[#, patt]& /@ p

pullVectors[a_^n_Integer, patt_] :=
  pullVectors[a, patt] pullVectors[a^(n-1), patt] /; n > 1

pullVectors[Sqr[p_], patt_] := Block[{i=Unique[]}, p[i] p[i]] /;
  MatchQ[p, patt] && HEPTensorSignature[p] === {Lorentz} 

pullVectors[SDot[p_, q_], patt_] := Block[{i=Unique[]}, p[i] q[i]] /;
  (MatchQ[p, patt] || MatchQ[q, patt]) &&
  HEPTensorSignature[p] === HEPTensorSignature[q] === {Lorentz}

pullVectors[Gs[p_], patt_] := Block[{i = Unique[]}, Ga[i] p[i]] /;
  MatchQ[p, patt]

pullVectors[HoldPattern[EpsDot[q1___, p_, q2___][mm__]], patt_] :=
  Block[{mu=Unique[]},
    (-1)^Length[{q2}] pullVectors[EpsDot[q1, q2][mu, mm], patt] p[mu]] /;
  MatchQ[p, patt]

pullVectors[HoldPattern[EpsDot[q1___, p_, q2___]], patt_] := Block[{mu=Unique[]},
  (-1)^Length[{q2}] pullVectors[EpsDot[q1, q2][mu], patt] p[mu]] /;
  MatchQ[p, patt]

pullVectors[(h_?LinearQ)[aa___], patt_] := h @@ (pullVectors[#, patt]& /@ {aa})

pullVectors[x_, patt_] := x

PullVectors[ex_, patt_] := Block[{res = ex},
  res = HEPExpand[ex, patt]; 
  res = pullVectors[res, patt];
  PullScalars[res, patt]
]

PullVectors[ex_] := PullVectors[ex, _]


(* ToComponents *)

ToComponents[ex_, rul_List, OptionsPattern[]] := Block[{vecs, vpatt},
  vecs = First /@ rul;
  Scan[If[HEPTensorSignature[#] =!= {Lorentz},
          Message[ToComponents::novec, #]]&, vecs]; 
  vpatt = Alternatives @@ vecs;
  ex /. {
    SDot[p:vpatt, q:vpatt] :>
      OptionValue[PostProcess][MProd[p /. rul, q /. rul]],
    Sqr[p:vpatt] :>
      OptionValue[PostProcess][MProd[p /. rul, p /. rul]],
    EpsDot[p1:vpatt, p2:vpatt, p3:vpatt, p4:vpatt] :>
      OptionValue[PostProcess][Det[{p1, p2, p3, p4} /. rul]]
  }
]

ToComponents[ex_, rul_, p:OptionsPattern[]] := ToComponents[ex, {rul}, p]


(* CalcDiracTraces *)

CalcDiracTraces[ex_, OptionsPattern[]] := Block[{result=ex, times},
  (* substitute shorthands *)
  result = result /. t_DiracTr :> contractme[DiracContract[HEPExpand[
    DiracSubstitute[ExplicitSlash[t]], GI|G5|_Ga]]];

  (* mark 4D indices *)
  result = result /. {
    EtaHat[mu_, tilde[nu_]] -> 0,
    EtaHat[tilde[nu_], mu_] -> 0
  };
  result = result /. Times -> times;
  result = result //. 
    times[a___, (p_?FourDimensionalQ)[mu_], b___] :>
      (times[a, p[mu], b] /. mu -> tilde[mu]) /; Head[mu] =!= tilde;
  result = result //. {
    times[a___, HoldPattern[Eta[tilde[mu_], nu_]], b___] :>
      (times[a, Eta[tilde[mu], nu], b] /. nu -> tilde[nu]) /; Head[nu] =!= tilde,
    times[a___, HoldPattern[Eta[nu_, tilde[mu_]]], b___] :>
      (times[a, Eta[nu, tilde[mu]], b] /. nu -> tilde[nu]) /; Head[nu] =!= tilde
  };
  result = result /. times -> Times;

  (* calculate traces *)
  antiCommutingG5 = OptionValue[AntiCommutingG5];
  result = result /. DiracTr[a_Dot] :>
    tr @@ (List @@ a /. { Ga -> Identity, G5 -> g5 });

  (* cleanup *)
  result = result //. {
    Eta[hat[mu_], nu_] :> EtaHat[mu, nu],
    Eta[mu_, hat[nu_]] :> EtaHat[mu, nu],
    Eps[a___,hat[mu_],b___] -> 0
  };
  result = result /. hat -> Identity;
  result = result /. tilde -> Identity;
  result = result /. contractme[x_] :>
    LorentzContract[Expand[x, Eta|EtaHat|Eps]];

  result
]


(* tr definitions *)

(* traces with odd number of gamma matrices *)

tr[a___] := 0 /; OddQ[Length[{a}]-Count[{a},g5]]

(* reduction to terms with only one g5 *)

tr[a___,g5,b___,c:Except[g5]] := tr[c,a,g5,b]

tr[a___,g5,g5,b___] := tr[a,b]

tr[a___,g5,tilde[b_],c___] := -tr[a,tilde[b],g5,c];

tr[a___,g5,mu_,c___] := If[antiCommutingG5 === True,
  -tr[a,mu,g5,c],
  -tr[a,mu,g5,c] + 2 tr[a,hat[mu],g5,c]];

(* traces with up to 4 gamma matrices *)

tr[] := 4
 
tr[g5] := 0
 
tr[a_, b_, g5] := 0
 
tr[a_, b_] := 4 Eta[a,b]
 
tr[a_, b_, c_, e_, g5] := 4I Eps[a, b, c, e]
 
tr[a_, b_, c_, e_] :=
  4 (Eta[a,b] Eta[c,e] - Eta[a,c] Eta[b,e] + Eta[a,e] Eta[b,c])

(* contraction of two gamma matrices in one trace *)

tr[a___,hat[mu_],b___,mu_,c___] := tr[a, hat[mu], b, hat[mu], c]

tr[a___,mu_,b___,hat[mu_],c___] := tr[a, hat[mu], b, hat[mu], c]

tr[a___, mu_, mu_, b___] := Switch[Head[mu],
  hat, (Dim-4) tr[a, b],
  _  , Dim tr[a, b]
]
 
tr[a___, mu_, b_, mu_, c___] := Switch[Head[mu],
  hat, (4-Dim) tr[a,b,c] + 2 tr[a,hat[b],c],
  _  , (2-Dim) tr[a,b,c]
]
 
tr[a___, mu_, b_, c_, mu_, e___] := Switch[Head[mu],
  hat, (Dim-4) tr[a,b,c,e]-2 tr[a,hat[b],c,e]+2 tr[a,hat[c],b,e],
  _  , 4 Eta[b,c] tr[a, e] + (Dim-4) tr[a, b, c, e]
]
 
tr[a___, mu_, b_, c_, e_, mu_, f___] := Switch[Head[mu],
  hat, (4-Dim) tr[a,b,c,e,f] + 2 tr[a,hat[b],c,e,f] -
       2 tr[a,hat[c],b,e,f] + 2 tr[a,hat[e],b,c,f],
  _  , (4-Dim) tr[a,b,c,e,f] - 2 tr[a, e, c, b, f]
]
 
tr[a___, mu_, b__, mu_, c___] := Switch[Head[mu],
  hat, (-1)^Length[{b}] ( (Dim-4) tr[a,b,c] +
         Block[{i},
               2 Sum[(-1)^i tr @@ Flatten[ {{a}, Prepend[Drop[{b}, {i}],
                                            hat[ {b}[[i]]] ],{c} },1 ],
                     {i,1,Length[{b}]}]
         ] ),
  _  , (-1)^Length[{b}] ( (Dim-4) tr[a,b,c] +
         2 tr @@ Flatten[{{a},Join[Reverse[Drop[{b},3-Length[{b}]]],
                         Drop[{b},3]], {c} }, 1] +
         Block[{i},
               2 Sum[(-1)^i tr @@ Flatten[{{a}, Prepend[Drop[{b}, {i}],
                                           {b}[[i]]], {c} },1 ],
                     {i,4,Length[{b}]}]
         ] )
]

(* traces of more than 4 gamma matrices *)

tr[a__, g5] :=
  Block[{i},
    Sum[(-1)^i Eta[{a}[[1]],{a}[[i]]] *
    tr @@ Append[Rest[Drop[{a},{i}]], g5], {i,2,Length[{a}]}] ] -
  Block[{i1=Unique[],i2=Unique[],i3=Unique[]},
    I/6 Eps[{a}[[1]],i1,i2,i3] tr @@ Join[Rest[{a}],{i1,i2,i3}]]
 
tr[a__] := Block[{i},
  Sum[(-1)^i Eta[{a}[[1]],{a}[[i]]] * tr @@ Rest[Drop[{a},{i}]],
      {i,2,Length[{a}]}]
]


(* PolarizationSums *)

FermionPolarizationSums[ex_, rul_List] := Block[
  {mass, pol, coeff, patt},

  mass[p_, r_] := Block[{m = p /. r},
    If[tensorQ[m], Print[m]; Message[VectorPolarizationSums::invm, p]; Abort[]];
    m
  ];

  patt = Alternatives @@ (First /@ rul);
  {pol,coeff} = HEPCollectListed[ex, USp|VSp];

  pol = pol //. {
    HoldPattern[USp[p_][s_,al_] Bar[USp[p_]][s_,bt_]] :>
      (Gs[p] + GI mass[p, rul])[al,bt] /; MatchQ[p, patt],
    HoldPattern[(A_ . USp[p_][s_]) (Bar[USp[p_][s_]] . B_)] :>
      A.(Gs[p] + GI mass[p, rul]).B /; MatchQ[p, patt],
    HoldPattern[Dot[Bar[USp[p_]][s_], M___, USp[p_][s_]]] :>
      DiracTr[Dot[(Gs[p] + GI mass[p, rul]), M]] /; MatchQ[p, patt],
    HoldPattern[VSp[p_][s_,al_] Bar[VSp[p_]][s_,bt_]] :>
      (Gs[p] - GI mass[p, rul])[al,bt] /; MatchQ[p, patt],
    HoldPattern[(A_ . VSp[p_][s_]) (Bar[VSp[p_]][s_] . B_)] :>
      A.(Gs[p] - GI mass[p, rul]).B /; MatchQ[p, patt],
    HoldPattern[Dot[Bar[VSp[p_]][s_], M___, VSp[p_][s_]]] :>
      DiracTr[Dot[(Gs[p] - GI mass[p, rul]), M]] /; MatchQ[p, patt]
  };

  pol.coeff
]

FermionPolarizationSums[ex_, rul:(_Rule|_RuleDelayed)] :=
  FermionPolarizationSums[ex, {rul}]


VectorPolarizationSums[ex_, rul_List, OptionsPattern[]] := Block[
  {pol,coeff,mass,nvec,patt},

  mass[p_, r_] := Block[{m = p /. r},
    If[tensorQ[m], Message[VectorPolarizationSums::invm, p]; Abort[]];
    m
  ];

  nvec[p_, r_] := Block[{n = p /. r},
    If[HEPTensorSignature[n] =!= {Lorentz},
       Message[VectorPolarizationSums::invn, p]; Abort[]];
    n
  ];

  {pol,coeff} = HEPCollectListed[ex, Pol|MPol];
  patt = Alternatives @@ (First /@ rul);

  pol = PullVectors[#, Pol[_][_]|CC[Pol[_]][_]|MPol[_][_]|CC[MPol[_]][_]]& /@
    pol;
  pol = pol /. {
    Pol[k:patt][lm_][mu_] -> Pol[k][lm, mu],
    CC[Pol[k:patt]][lm_][mu_] -> CC[Pol[k]][lm, mu],
    MPol[k:patt][lm_][mu_] -> MPol[k][lm, mu],
    CC[MPol[k:patt]][lm_][mu_] -> CC[MPol[k]][lm, mu]
  };
  pol = pol //. {
    CC[Pol[k:patt]][lm_, mu_] Pol[k:patt][lm_, nu_] :> Block[{n},
      n=nvec[k,rul];
      -Eta[mu,nu] - Sqr[n] k[mu] k[nu]/
        OptionValue[PostProcess][SDot[k,n]^2] +
      (n[mu] k[nu] + k[mu] n[nu])/
        OptionValue[PostProcess][SDot[k,n]]
    ],
    CC[MPol[k:patt]][lm_, mu_] MPol[k:patt][lm_, nu_] :> 
      -Eta[mu,nu] + k[mu] k[nu]/mass[k,rul]^2
  };

  pol.coeff
]

VectorPolarizationSums[ex_, rul:(_Rule|_RuleDelayed), opt:OptionsPattern[]] :=
  VectorPolarizationSums[ex, {rul}, opt]


End[]

EndPackage[]


