BeginPackage["HEPMath`PaVe`", {"HEPMath`Tensors`", "HEPMath`Lorentz`"}]

PaVe::usage = "PaVe[n, mu, lam][inds][args] denotes the n-point one-loop
 tensor integral with renormalization scale mu, IR regulator lam, indices
 inds and arguments args. See the LoopTools documentation for details and
 conventions."

DPaVe::usage = "DPaVe[2, mu, lam][inds][psq, m0sq, m1sq] denotes the derivative
 (with respect to psq) of the 2-point one-loop tensor integral with
renormalization scale mu, IR regulator lam, indices inds and kinetic arguments
psq, m0sq, m1sq. See the LoopTools documentation for details and conventions."

Den::usage = "Den[q, m] denotes the propagator denominator 1/(SDot[q,q]-m^2)."

PaVeIntegrate::usage = "PaVeIntegrate[expr, l, mu] integrates expr over the loop
momentum l by replacing occurrences of (numerator) loop momenta l and propagator
denominators Den[q,m] with one-loop tensor functions PaVe, using the
renormalization scale mu."

PaVeIntegrate::noint = "Cannot integrate `1`"

IRDivergentParts::usage = "IRDivergentParts is an option of PaVeIntegrate. If
set to True, IR divergent parts of the one-loop integrals are included as
a power series in IRDiv. Defaults to False."

LoopFactor::usage = "LoopFactor is an option of PaVeIntegrate. All one-loop
integrals are mutiplied with this factor. The default setting is I Pi^2, which
matches the default setting of the PreFactor option of CreateFeynAmp."

DenOrder::usage = "DenOrder is an option of PaVeIntegrate which can be used to
define an ordering for the denominators. Possible values are ordering functions
similar to the second argument of Sort."

Options[PaVeIntegrate] = {
  IRDivergentParts -> False,
  LoopFactor -> I Pi^2,
  DenOrder -> (Signature[{#1,#2}] === 1 &)
}

Div::usage = "Div denotes the UV divergence."

IRDiv::usage = "IRDiv denotes the IR divergence."

ApplyToPaVeArgs::usage = "ApplyToPaVeArgs[f, expr] applies function f to the
arguments of all arguments of all tensor integrals in expr."

DenReduce::usage = "DenReduce[expr, l] reduces the number of propagator
denominators with loop momentum l in expr by cancelling them against scalar
products in the numerator."


Begin["`Private`"]

pave::usage = "Interal symbol for one-loop integrals"

DeclareRealHolomorphic[Den]

DeclareReal[Div]

DeclareReal[IRDiv]

PaVe[nn__][ii__] := PaVe[nn]@@Sort[{ii}] /; !OrderedQ[{ii}]

DPaVe[nn__][ii__] := PaVe[nn]@@Sort[{ii}] /; !OrderedQ[{ii}]

PaVe[_, _, k_Integer][__][__] := 0 /; k > 2

DPaVe[_, _, k_Integer][__][__] := 0 /; k > 2


aargs[{Den[l_, m_]}, l_] := {m^2}

bargs[{Den[l_, m0_], Den[l_+p_, m1_]}, l_] := {Sqr[p], m0^2, m1^2}

cargs[{Den[l_, m0_], Den[l_+k1_, m1_], Den[l_+k2_, m2_]}, l_] :=
  {Sqr[k1], Sqr[k2-k1], Sqr[k2], m0^2, m1^2, m2^2}

dargs[{Den[l_, m0_], Den[l_+k1_, m1_], Den[l_+k2_, m2_],
       Den[l_+k3_, m3_]}, l_] :=
  {Sqr[k1], Sqr[k2-k1], Sqr[k3-k2], Sqr[k3], Sqr[k2], Sqr[k3-k1],
   m0^2, m1^2, m2^2, m3^2}


unZipMatches[l_List, patt_] := Block[{true={},false={}},
  Scan[If[MatchQ[#, patt]===True, AppendTo[true, #], AppendTo[false, #]]&, l];
  {true,false}
]

paveExpand[ex_, l_, ord_] := Block[{res=ex},
  res = res /. pave[ll_, dd_] :> pave[HEPExpand[#, l]& /@ ll, dd];
  res = res //. pave[{a___,s_Plus,b___}, dd_] :> (pave[{a,#,b}, dd]& /@ s);
  res = res //. pave[{a___,(c_ /; FreeQ[k, l]) k_,b___}, dd_] :>
    c pave[{a,k,b}, dd];
  res = res //. pave[{a___,(k_ /; FreeQ[k, l]),b___}, dd_] :> k pave[{a,b}, dd];
  res = res /. pave[ll_, dd_] :> pave[Sort[ll], Sort[dd, ord]];
  res = res /. pave[ll_, {d1__, Den[l, m_], d2___}] :>
    pave[ll, {Den[l, m], d1, d2}];
  HEPCollect[res, pave]
]

paveShift[ex_, l_, ord_] := Block[{res=ex},
  res = res /. pave[ll_, {Den[l+q_, m_], dd___}] :>
    (pave[ll, {Den[l+q, m], dd}] /. l -> l-q) /; FreeQ[{dd}, Den[l, _]];
  paveExpand[res, l, ord]
]

paveReduce[ex_, l_, ord_] := Block[{res=ex},
  res = res /. pave[{a___, l[mu_]^2, b___}, {d1___, Den[l, m_], d2___}] :>
    pave[{a,b},{d1, d2}] + m^2 pave[{a,b}, {d1, Den[l, m], d2}];
  paveShift[res, l, ord]
]

PaVeIntegrate[ex_, l_, ren_, OptionsPattern[]] := Block[
  {int, coeff, lam, res, fac},

  fac = OptionValue[LoopFactor];

  res = PullVectors[ex, l];
  res = res /. {
    Den[-l+q_, m_] :> Den[Expand[l-q], m],
    Den[-l, m_] -> Den[l, m]
  };

  {int, coeff} = HEPCollectListed[res, l];
  int = If[Head[#] === Times, List@@#, {#}]& /@ int;
  int = (pave @@ unZipMatches[#,Except[_Den]])& /@ int;
  int = int /. pave[ll_, dd_] :> pave[Sort[ll], Sort[dd, OptionValue[DenOrder]]];
  res = int.coeff;
  res = paveShift[res, l, OptionValue[DenOrder]];
  res = FixedPoint[paveReduce[#, l, OptionValue[DenOrder]]&, res];

  (* substitute tensor integrals *)
  {int, coeff} = HEPCollectListed[res, l];
  int = int /. {
    pave[{}, d:{Den[l,m_]}] :> fac PaVe[1,ren,lam][0]@@aargs[d,l],
    pave[{l[_]}, {Den[l,m_]}] :> 0,
    pave[{l[mu_], l[nu_]}, d:{Den[l,m_]}] :>
      fac PaVe[1,ren,lam][0,0]@@aargs[d,l] Eta[mu, nu],

    pave[{}, d:{Den[l,m0_], Den[l+p_,m1_]}] :>
      fac PaVe[2,ren,lam][0]@@bargs[d,l],
    pave[{l[mu_]}, d:{Den[l,m0_], Den[l+p_,m1_]}] :>
      fac PaVe[2,ren,lam][1]@@bargs[d,l],
    pave[{l[mu_], l[nu_]}, d:{Den[l,m0_], Den[l+p_,m1_]}] :> fac (
      PaVe[2,ren,lam][0,0]@@bargs[d,l] Eta[mu,nu] +
      PaVe[2,ren,lam][1,1]@@bargs[d,l] p[mu] p[nu]),

    pave[{}, d:{Den[l,m0_], Den[l+k1_, m1_], Den[l+k2_, m2_]}] :>
      fac PaVe[3,ren,lam][0]@@cargs[d,l],
    pave[{l[mu_]}, d:{Den[l,m0_], Den[l+k1_, m1_], Den[l+k2_, m2_]}] :> fac (
      PaVe[3,ren,lam][1]@@cargs[d,l] k1[mu] +
      PaVe[3,ren,lam][2]@@cargs[d,l] k2[mu]),
    pave[{l[mu_], l[nu_]}, d:{Den[l,m0_], Den[l+k1_, m1_], Den[l+k2_, m2_]}] :>
      fac Block[{args=cargs[d,l], kk={k1, k2}},
        PaVe[3,ren,lam][0,0]@@args Eta[mu,nu] +
        Sum[PaVe[3,ren,lam][i,j]@@args kk[[i]][mu] kk[[j]][nu],
            {i,1,2}, {j,1,2}]],
    pave[{l[mu_], l[nu_], l[ro_]},
         d:{Den[l,m0_], Den[l+k1_, m1_], Den[l+k2_, m2_]}] :> fac *
      Block[{args=cargs[d,l], kk={k1, k2}},
        Sum[PaVe[3,ren,lam][0,0,i]@@args (Eta[mu, nu] kk[[i]][ro] +
                               Eta[nu, ro] kk[[i]][mu] +
                               Eta[ro, mu] kk[[i]][nu]), {i,1,2}] +
        Sum[PaVe[3,ren,lam][i,j,k]@@args kk[[i]][mu] kk[[j]][nu] kk[[k]][ro],
            {i,1,2}, {j,1,2}, {k,1,2}]],

    pave[{},
         d:{Den[l,m0_], Den[l+k1_, m1_], Den[l+k2_, m2_], Den[l+k3_, m3_]}] :>
      fac PaVe[4,ren,lam][0]@@dargs[d,l],
    pave[{l[mu_]},
         d:{Den[l,m0_], Den[l+k1_, m1_], Den[l+k2_, m2_], Den[l+k3_, m3_]}] :>
      fac Block[{args=dargs[d,l], kk={k1,k2,k3}},
        Sum[PaVe[4,ren,lam][i]@@args kk[[i]][mu], {i,1,3}]],
    pave[{l[mu_], l[nu_]},
         d:{Den[l,m0_], Den[l+k1_, m1_], Den[l+k2_, m2_], Den[l+k3_, m3_]}] :>
      fac Block[{args=dargs[d,l], kk={k1,k2,k3}},
        PaVe[4,ren,lam][0,0]@@args Eta[mu,nu] +
        Sum[PaVe[4,ren,lam][i,j]@@args kk[[i]][mu] kk[[j]][nu],
            {i,1,3}, {j,1,3}]],
    pave[{l[mu_], l[nu_], l[ro_]},
         d:{Den[l,m0_], Den[l+k1_, m1_], Den[l+k2_, m2_], Den[l+k3_, m3_]}] :>
      fac Block[{args=dargs[d,l], kk={k1,k2,k3}},
        Sum[PaVe[4,ren,lam][0,0,i]@@args (Eta[mu,nu] kk[[i]][ro] +
                               Eta[ro,mu] kk[[i]][nu] +
                               Eta[nu,ro] kk[[i]][mu]), {i,1,3}] +
        Sum[PaVe[4,ren,lam][i,j,k]@@args kk[[i]][mu] kk[[j]][nu] kk[[k]][ro],
            {i,1,3}, {j,1,3}, {k,1,3}]],
    pave[{l[mu_], l[nu_], l[ro_], l[sg_]},
         d:{Den[l,m0_], Den[l+k1_, m1_], Den[l+k2_, m2_], Den[l+k3_, m3_]}] :>
      fac Block[{args=dargs[d,l], kk={k1,k2,k3}},
        PaVe[4,ren,lam][0,0,0,0]@@args (Eta[mu,nu] Eta[ro,sg] +
                             Eta[mu,ro] Eta[nu,sg] +
                             Eta[mu,sg] Eta[nu,ro])+
        Sum[PaVe[4,ren,lam][0,0,i,j]@@args (Eta[mu,nu] kk[[i]][ro] kk[[j]][sg] +
                                 Eta[mu,ro] kk[[i]][nu] kk[[j]][sg] +
                                 Eta[mu,sg] kk[[i]][nu] kk[[j]][ro] +
                                 Eta[nu,ro] kk[[i]][mu] kk[[j]][sg] +
                                 Eta[nu,sg] kk[[i]][mu] kk[[j]][ro] +
                                 Eta[ro,sg] kk[[i]][mu] kk[[j]][nu]),
            {i,1,3}, {j,1,3}] +
        Sum[PaVe[4,ren,lam][i,j,k,l]@@args *
              kk[[i]][mu] kk[[j]][nu] kk[[k]][ro] kk[[l]][sg],
            {i,1,3}, {j,1,3}, {k,1,3}, {l,1,3}]]
  };
  int = int /. pave[{ll___}, {dd__}] :>
    (Message[PaVeIntegrate::noint, Times@@{ll,dd}]; Times@@{ll,dd});

  (* make divergent parts explicit *)
  int = int /. {
    PaVe[1,ren,lam][0][msq_] -> msq Div + PaVe[1,ren,lam][0][msq],
    PaVe[1,ren,lam][0,0][msq_] -> msq^2/4 Div + PaVe[1,ren,lam][0,0][msq],

    PaVe[2,ren,lam][0][psq_, m0sq_, m1sq_] ->
      Div + PaVe[2,ren,lam][0][psq, m0sq, m1sq],
    PaVe[2,ren,lam][1][psq_, m0sq_, m1sq_] ->
      -Div/2 + PaVe[2,ren,lam][1][psq, m0sq, m1sq],
    PaVe[2,ren,lam][0,0][psq_, m0sq_, m1sq_] ->
      ((m0sq + m1sq)/2 - psq/12) Div + PaVe[2,ren,lam][0,0][psq, m0sq, m1sq],
    PaVe[2,ren,lam][1,1][psq_, m0sq_, m1sq_] ->
      Div/3 + PaVe[2,ren,lam][1,1][psq, m0sq, m1sq],

    PaVe[3,ren,lam][0,0][aa__] -> Div/4 + PaVe[3,ren,lam][0,0][aa],
    PaVe[3,ren,lam][0,0,i_][aa__] -> -Div/12 + PaVe[3,ren,lam][0,0,i][aa],
    PaVe[3,ren,lam][0,0,0,0][p1sq_, p2sq_, p3sq_, m0sq_, m1sq_, m2sq_] ->
      ((m0sq+m1sq+m2sq)/24 - (p1sq+p2sq+p3sq)/96) Div + 
      PaVe[3,ren,lam][0,0,0,0][p1sq, p2sq, p3sq, m0sq, m1sq, m2sq],
    PaVe[3,ren,lam][0,0,i_,i_][aa__] -> Div/24 + PaVe[3,ren,lam][0,0,i,i][aa],
    PaVe[3,ren,lam][0,0,i_,j_][aa__] -> Div/48 + PaVe[3,ren,lam][0,0,i,j][aa],
    
    PaVe[4,ren,lam][0,0,0,0][aa__] -> Div/24 + PaVe[4,ren,lam][0,0,0,0][aa]
  };

  (* add IR divergent parts *)
  If[OptionValue[IRDivergentParts],
    int = int /. PaVe[n_,mu_,lam][ii__][aa__] ->
      PaVe[n,mu,0][ii][aa] + IRDiv PaVe[n,mu,1][ii][aa] +
      IRDiv^2 PaVe[n,mu,2][ii][aa]
    ,
    int = int /. PaVe[n_,mu_,lam] -> PaVe[n,mu,0]
  ];

  (* remove pave[{},{}] *)
  coeff = coeff /. pave[{},{}] -> 1;

  int.coeff
]

ApplyToPaVeArgs[f_, ex_] := ex /. PaVe[nn__][j__][aa__] :>
  PaVe[nn][j] @@ (f /@ {aa})

End[]

EndPackage[]


