Needs["HEPMath`"]

Needs["HEPMath`Compile`"]

(* Declare vectors *)
DeclareLorentzVectors[p, q, k, l]

(* Compile a function with two arguments.
   You can use HEPEvaluate to substitute evaluations in the function body. *)
expr = x^2 + y^2;
cf1 = HEPCompile[{{x, _Real}, {y, _Real}}, HEPEvaluate[expr] + 1.0];

(* To use expressions involving PaVe, Sqr, SDot or EpsDot in a compiled function
   you have to wrap HEPPrepare around them *)

b0 = PaVe[2, 1.0, 0][0][Sqr[p], m^2, m^2];
cf2 = HEPCompile[{{p, _Real, 1}, {m, _Real}}, HEPPrepare[b0]];

cf3 = HEPCompile[{{p, _Real, 1}, {q, _Real, 1}, {k, _Real, 1}, {l, _Real, 1}},
  HEPPrepare[EpsDot[p, q, k, l]]]

(* Build directory. *)
builddir = "build"

(* Name of the library. *)
libname = "myfuncs"

(* Create the build directory if it doesn't exist *)
If[!DirectoryQ[builddir], CreateDirectory[builddir]]

(* Generate code from compiled functions.
   "myfuncs" will be the name of the library. *)
modules = HEPCodeGenerate[{cf1, cf2, cf3}, {"f1", "f2", "f3"}, libname,
  TargetDirectory->builddir];


(* Compile the shared library. 
   To use the library in Python you have to put an underscore
   in front of its name. Also note that you need to link the Wolfram runtime
   library and dl *)

Needs["CCompilerDriver`"]

CreateLibrary[modules, "_"<>libname,
  "TargetDirectory" -> builddir,
  "Debug"->True, "Libraries"->{"WolframRTL", "dl"},
  "ShellCommandFunction"->Print]

(* Remove temporary files. Mathematica should do this but it doesn't *)
DeleteDirectory[#, DeleteContents->True]& /@
  FileNames[FileNameJoin[{builddir, "Working*"}]]

